package gearman

import (
	"sync"
	"testing"
	"time"
)

func TestSyncFIFO(t *testing.T) {
	var f SyncMessageFIFO
	f.Start()
	defer f.Close()
	if f.Get() != nil {
		t.Error("empty FIFO returned non nil")
	}
	f.Push(&Message{Type: MessageCommandMasterTX, String: "stset"})
	m := f.Get()
	if m == nil || m.Type != MessageCommandMasterTX || m.String != "stset" {
		t.Error("FIFO returend something strange")
	}

	var wg sync.WaitGroup
	wg.Add(100)
	for i := 0; i < 100; i++ {
		go func() {
			f.Push(&Message{Type: MessageCommandMasterTX, String: "stset"})
			wg.Done()
		}()
	}

	wg.Wait()
	time.Sleep(10 * time.Millisecond)
	ms := f.GetAll(200)
	if len(ms) != 100 {
		t.Error("returned a different amount of elements", len(ms))
	}

	wg.Add(100)
	for i := 0; i < 100; i++ {
		go func() {
			f.Push(&Message{Type: MessageCommandMasterTX, String: "stset"})
			wg.Done()
		}()
	}

	wg.Wait()
	time.Sleep(10 * time.Millisecond)
	ms = f.GetAll(50)
	if len(ms) != 50 {
		t.Error("returned a different amount of elements", len(ms))
	}
}

func TestSyncFIFONotify(t *testing.T) {
	var f SyncMessageFIFO
	f.Start()
	defer f.Close()

	var wg sync.WaitGroup
	wg.Add(100)
	for i := 0; i < 100; i++ {
		go func() {
			f.Push(&Message{Type: MessageCommandMasterTX, String: "stset"})
			wg.Done()
		}()
	}

	numMsg := 0
	for numMsg < 100 {
		select {
		case <-f.Notify():
			msg := f.Get()
			if msg != nil {
				numMsg++
			}
		}
	}
	wg.Wait()
}

func TestSyncFIFOFrontBack(t *testing.T) {
	var f SyncMessageFIFO
	f.Start()
	defer f.Close()

	f.PushAll([]*Message{{Type: MessageCommandSatTX, String: "back"}}, false)
	f.PushAll([]*Message{{Type: MessageCommandSatTX, String: "front"}}, true)
	f.PushAll([]*Message{{Type: MessageCommandSatTX, String: "back2"}}, false)
	f.PushAll([]*Message{{Type: MessageCommandSatTX, String: "front2"}}, true)

	msgs := make([]*Message, 0)
	for len(msgs) < 4 {
		select {
		case <-f.Notify():
			msg := f.Get()
			if msg != nil {
				msgs = append(msgs, msg)
			}
		}
	}
	if msgs[0].String != "front2" || msgs[1].String != "front" || msgs[2].String != "back" || msgs[3].String != "back2" {
		t.Error("Wrong order of msgs")
	}
}

func TestSyncFIFOFrontBack2(t *testing.T) {
	var f SyncMessageFIFO
	f.Start()
	defer f.Close()

	f.PushAll([]*Message{{Type: MessageCommandSatTX, String: "front"}}, true)
	f.PushAll([]*Message{{Type: MessageCommandSatTX, String: "front2"}}, true)
	f.PushAll([]*Message{{Type: MessageCommandSatTX, String: "back"}}, false)
	f.PushAll([]*Message{{Type: MessageCommandSatTX, String: "back2"}}, false)

	msgs := make([]*Message, 0)
	for len(msgs) < 4 {
		select {
		case <-f.Notify():
			msg := f.Get()
			if msg != nil {
				msgs = append(msgs, msg)
			}
		}
	}
	if msgs[0].String != "front2" || msgs[1].String != "front" || msgs[2].String != "back" || msgs[3].String != "back2" {
		t.Error("Wrong order of msgs")
	}
}

func TestSyncFIFODirect(t *testing.T) {
	var f SyncMessageFIFO
	f.Start()
	defer f.Close()

	var wg sync.WaitGroup
	wg.Add(100)
	for i := 0; i < 100; i++ {
		go func() {
			f.Direct() <- &Message{Type: MessageCommandMasterTX, String: "stset"}
			wg.Done()
		}()
	}

	numMsg := 0
	for numMsg < 100 {
		select {
		case <-f.Notify():
			msg := f.Get()
			if msg != nil {
				numMsg++
			}
		}
	}
	wg.Wait()
}

func BenchmarkSyncFIFO(b *testing.B) {
	var wg sync.WaitGroup

	var f SyncMessageFIFO
	f.Start()

	wg.Add(1)
	go func() {
		numMsg := 0
		for numMsg < 200000 {
			select {
			case <-f.Notify():
				msg := f.Get()
				if msg != nil {
					numMsg++
				}
			}
		}
		wg.Done()
	}()

	msg := &Message{}

	for i := 0; i < 200000; i++ {
		f.Push(msg)
		if i%20000 == 0 {
			b.Log("num messages in buffer:", f.messages.Len())
		}
	}

	wg.Wait()
	f.Close()
}
